/*
 *    for STM32F746 Discovery
 *
 *    Copyright (C) by Hasada & FCT Embedded System Research Circle
 */

#include "lcd_param.h"
#include "imgpro_api.h"
#include <string.h>
#include <stdlib.h>

typedef  unsigned short int  uint16_t;


char *bg_fname_list[] = {
  "bg320x240_1.bmp",
  "bg320x240_2.bmp",
  "bg320x240_3.bmp",
  "bg320x240_4.bmp",
  "bg320x240_5.bmp",
  "bg320x240_6.bmp",
  "bg320x240_7.bmp",
  "bg320x240_8.bmp",
  "bg320x240_9.bmp",
  "bg320x240_10.bmp",
  0x0
};

#define  DEF_FNAME_HEADER_0   'U'  /* unknown */
#define  DEF_FNAME_HEADER_1   'P'  /* photo */
#define  DEF_FNAME_HEADER_2   'O'  /* output */

char  fname_hedaer_table[] = {
  DEF_FNAME_HEADER_0,
  DEF_FNAME_HEADER_1,
  DEF_FNAME_HEADER_2,
};


void  make_filename( char header, int cnt, char *filename )
{
  int  i;
  char  *pnt;

  *filename++ = header;

  filename += 5;
  pnt = filename++;
  for( i=5 ; i>=0 ; i-- ) {
    *pnt-- = '0' + (cnt % 10);
    cnt = cnt / 10;
  }

  *filename++ = '.';
  *filename++ = 'b';
  *filename++ = 'm';
  *filename++ = 'p';
  *filename++ = 0x00;
}


/*
 *  photo : 320x240pixel RGB565
 *  size : unused
 */
void  output_bmp_file( unsigned char *photo, int size, int target )
{
  int  num;
  int  ret;
  char  tmp_fname[32];

  /* check output target */
  if( target < 1 && 2 < target ) {
    target = 0;
  }

  num = 0;
  do {
    make_filename( fname_hedaer_table[target], num, &tmp_fname[0] );
    ret = api_sdcard_file_check( tmp_fname );
    if( ret < 0 ) {
      /* sdcard access error */
      return;
    }
    num++;

    /* check save space */
    if( num > 999999 ) {
      syslog( 3, "ERROR : make filename : ret=%d", ret );
      return;
    }
  } while( ret != 1 );

  syslog( 3, "access filename : %s", tmp_fname );

  api_write_bmp_file( photo, tmp_fname );
}


void usr_get_rgb( uint16_t src, char *r, char *g, char *b)
{
  char cp0 = (src >> 8) & 0xff;
  char cp1 = src & 0xff;
  *r = (cp0 & 0xf8) >> 3;
  *g = (((cp0 & 0x07) << 3) & 0x38) + ((cp1 >> 5) & 0x07);
  *b = cp1 & 0x1f;
  
  //Normalization
  *r = (*r << 3) | (0x07 & (*r >> 2));
  *g = (*g << 2) | (0x03 & (*g >> 4));
  *b = (*b << 3) | (0x07 & (*b >> 2));
  
  return;
}


void usr_set_rgb( char r, char g, char b, uint16_t *dst)
{
  uint16_t cp0 = 0;
  uint16_t cp1 = 0;
  uint16_t tmp_r = 0;
  uint16_t tmp_g = 0;
  uint16_t tmp_b = 0;
  
  r = r >> 3;
  g = g >> 2;
  b = b >> 3;
  
  tmp_r |= 0x00ff & r;
  tmp_g |= 0x00ff & g;
  tmp_b |= 0x00ff & b;
  
  cp0 = ((tmp_r << 3) & 0xF8) | ((tmp_g >> 3) & 0x07);
  cp1 = ((tmp_g << 5) & 0xE0) | (tmp_b & 0x1F);
  
  *dst = cp1 | (cp0 << 8);
}


void  usr_Diration( char* wrk, int width, int height) {
	
	int x, y;
	int xx, yy;
	int isWhite = 0;
	
	for (y = 0; y < height; y++) {
		for(x = 0 ; x < width ; x++) {
			
			if (wrk[y * width + x] > 0) {
				continue;
			}
			
			//check
			isWhite = 0;
			for (yy = y - 1; yy < y + 1; yy++) {
				for (xx = x - 1; xx < x + 1; xx++) {
					if (wrk[yy * width + xx] > 0) {
						wrk[y * width + x] = 255;
						isWhite = 1;
						break;
					}
				}
				if (isWhite != 0) {
					break;
				}
			}
		}
	}
}


void  usr_Erosion( char* wrk, int width, int height) {
	
	int x, y;
	int xx, yy;
	int isBlack = 0;
	
	for (y = 0; y < height; y++) {
		for(x = 0 ; x < width ; x++) {
			
			if (wrk[y * width + x] == 0) {
				continue;
			}
			
			//check
			isBlack = 0;
			for (yy = y - 1; yy < y + 1; yy++) {
				for (xx = x - 1; xx < x + 1; xx++) {
					if (wrk[yy * width + xx] == 0) {
						wrk[y * width + x] = 0;
						isBlack = 1;
						break;
					}
				}
				if (isBlack != 0) {
					break;
				}
			}
		}
	}
}


void  usr_copy_fb_to_v1_back_diff( void )
{
  int x,y;
  uint16_t  *src = (uint16_t *)DEF_LCD_FB_TOP;
  uint16_t  *dst = (uint16_t *)DEF_LCD_VUR_FB1;
  uint16_t  *background = (uint16_t *)DEF_LCD_VUR_FB1;
  char  *wrk = (char *)DEF_LCD_FB_WRK;
  uint16_t  *tmp_src = NULL;
  uint16_t  *tmp_back = NULL;
  char r, g, b;
  char back_r, back_g, back_b;
  float diffY;
  
	float cnt_w = 0;
	float cnt_b = 0;
	float sum_w = 0;
	float sum_b = 0;
	float max = 0;
	float val = 0;
	float max_th = 0;
	
	for (y = 0; y < DEF_LCD_FB_HEIGHT; y++) {
		for( x = 0 ; x < DEF_LCD_FB_WIDTH ; x++ ) {
			tmp_src = src + x + (y * DEF_LCD_FB_WIDTH);
			tmp_back = background + x + (y * DEF_LCD_FB_WIDTH);
			usr_get_rgb(*tmp_src, &r, &g, &b);
			usr_get_rgb(*tmp_back, &back_r, &back_g, &back_b);

			r = abs(r - back_r);
			g = abs(g - back_g);
			b = abs(b - back_b);
			diffY = 0.333 * r + 0.333 * g + 0.333 * b;
			
			//save absDiff (Gray Image)
			*wrk = diffY;
			wrk++;
			
		}
	}
	
	//Binarization (Otsu) threshold
	cnt_w = 0;
	cnt_b = 0;
	sum_w = 0;
	sum_b = 0;
	max = 0;
	val = 0;
	max_th = 0;
	for (int th = 1; th < 255; th++) {
		
		cnt_w = 0;
		cnt_b = 0;
		sum_w = 0;
		sum_b = 0;
		wrk = (char *)DEF_LCD_FB_WRK;
		
		for (y = 0; y < DEF_LCD_FB_HEIGHT; y++) {
			for( x = 0 ; x < DEF_LCD_FB_WIDTH ; x++ ) {
				if (*wrk >= th) {
					cnt_w++;
					sum_w += *wrk;
				} else {
					cnt_b++;
					sum_b += *wrk;
				}
				wrk++;
			}
			val = (cnt_b / 100) * (cnt_w / 100) * 
						((sum_w / cnt_w) - (sum_b / cnt_b)) * ((sum_w / cnt_w) - (sum_b / cnt_b));
			
			if (max < val) {
				max = val;
				max_th = th;
			}
		}
	}
	syslog(3, "th=%d", (int)max_th);
	
	//Binarization 
	wrk = (char *)DEF_LCD_FB_WRK;
	for (y = 0; y < DEF_LCD_FB_HEIGHT; y++) {
		for( x = 0 ; x < DEF_LCD_FB_WIDTH ; x++ ) {
			
			if (*wrk >= max_th) {
				*wrk = 255;
			} else {
				*wrk = 0;
			}
			wrk++;
		}
	}
	
	//Opening
	for (int i = 0; i < 2; i++) {
		usr_Erosion(wrk, DEF_LCD_FB_WIDTH, DEF_LCD_FB_HEIGHT);
	}
	for (int i = 0; i < 2; i++) {
		usr_Diration(wrk, DEF_LCD_FB_WIDTH, DEF_LCD_FB_HEIGHT);
	}
	
	//Closing
	for (int i = 0; i < 2; i++) {
		usr_Diration(wrk, DEF_LCD_FB_WIDTH, DEF_LCD_FB_HEIGHT);
	}
	for (int i = 0; i < 2; i++) {
		usr_Erosion(wrk, DEF_LCD_FB_WIDTH, DEF_LCD_FB_HEIGHT);
	}
	
	//output img if larger than threshold
	src = (uint16_t *)DEF_LCD_FB_TOP;
	dst = (uint16_t *)DEF_LCD_VUR_FB1;
	
	//copy new background img
	memcpy(dst, (void*)DEF_LCD_VUR_FB2, DEF_LCD_FB_SIZE);
	
	//draw foreground img
	wrk = (char *)DEF_LCD_FB_WRK;
	for (y = 0; y < DEF_LCD_FB_HEIGHT; y++) {
		for( x = 0 ; x < DEF_LCD_FB_WIDTH ; x++ ) {
			
			tmp_src = src + x + (y * DEF_LCD_FB_WIDTH);
			usr_get_rgb(*tmp_src, &r, &g, &b);
			
			if (*wrk > 0) {
				
				usr_set_rgb(r, g, b, dst);
				
			} else {
				
				//usr_set_rgb(0, 0, 0, dst);
				
			}
			wrk++;
			dst++;
		}
	}
}


void  func_key0( void )
{
    int  lcd;

	//Get Foreground Img from DEF_LCD_FB_TOP (Diff from current DEF_LCD_VUR_FB1)
	//Add new background img from DEF_LCD_VUR_FB2 (current selected image)
	//Disp to DEF_LCD_VUR_FB1 (Update)
    lcd = 1;
    usr_copy_fb_to_v1_back_diff();
    api_copy_v_to_r( lcd );
    api_disp_lcd( lcd );
}


void  func_key1( void )
{
    int  lcd;

    //Save Background Img to DEF_LCD_VUR_FB1
    lcd = 1;
    api_copy_fb_to_v( lcd );
    api_copy_v_to_r( lcd );
    api_disp_lcd( lcd );
}


void  func_key2( void )
{
    int  lcd;

	//Select Img from SD Card to DEF_LCD_VUR_FB2
    static  int  bg_pos = 0;

    lcd = 2;
    api_sdcard_read( bg_fname_list[bg_pos], (unsigned char *)DEF_LCD_USER_BMP1, DEF_LCD_BMP_SIZE );
    api_copy_bmp_to_v( (unsigned char *)DEF_LCD_USER_BMP1, lcd );
    api_copy_v_to_r( lcd );
    api_disp_lcd( lcd );

    bg_pos++;
    if( bg_fname_list[ bg_pos ] == 0x0 ) {
        bg_pos = 0;
    }
}


void  userfunc_main( int btnnum )
{
    /* stop camara input */
    api_camera_stop();

    switch( btnnum ) {
    case 1:
      func_key0();
      output_bmp_file( (unsigned char *)DEF_LCD_VUR_FB1, DEF_PHOTO_BYTE, 1 );
      break;
    case 2:
      func_key1();
      break;
    case 3:
      func_key2();
      break;
    }

    /* restart camera input */
    api_camera_start();
}


/*
 *  USER initalize function
 */
void  userfunc_init( void )
{
}

